/*
 * Autogenerated by read_ucd.py from The Unicode Standard 15.0.0
 *
 * String literals giving the currently supported version of Unicode.
 * Useful for error messages and 'about' boxes.
 */

#define UNICODE_VERSION_FULL "The Unicode Standard 15.0.0"
#define UNICODE_VERSION_SHORT "15.0.0"
